import os
import platform
import subprocess
from datetime import datetime, timedelta
from typing import IO

from config import FFMPEG_EXEC, VIDEO_BITRATE, FFMPEG_USE_NVIDIA_GPU, VIDEO_CLIP_EACH_SEC, VIDEO_CLIP_OVERFLOW_SEC, \
    FFMPEG_USE_INTEL_GPU


def base_ts_to_filename(start_ts: float, is_mp4=False) -> str:
    base_start = datetime.fromtimestamp(start_ts)
    if is_mp4:
        return base_start.strftime("%Y%m%d_%H%M.mp4")
    else:
        return base_start.strftime("%Y%m%d_%H%M.flv")


def get_video_real_duration(filename):
    ffmpeg_process = subprocess.Popen([
        FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-i", filename, "-c", "copy", "-f", "null", "-"
    ], stdout=subprocess.PIPE)
    return handle_ffmpeg_output(ffmpeg_process.stdout)


def multi_gpu_encode_video_with_subtitles(orig_filename: str, subtitles: list[str], base_ts: float):
    new_filename = base_ts_to_filename(base_ts)
    if not (FFMPEG_USE_NVIDIA_GPU and FFMPEG_USE_INTEL_GPU):
        print("[!]Please Enable Both Of Encode Acceleration")
        print("[!]Fallback to normal")
        encode_video_with_subtitles(orig_filename, subtitles, new_filename)
        return [new_filename]
    _duration_str = get_video_real_duration(orig_filename)
    duration = duration_str_to_float(_duration_str)
    if duration > (VIDEO_CLIP_EACH_SEC * 3):
        # qsv 压制前2段，剩余交由nvenc压制
        new_filename0 = base_ts_to_filename(base_ts)
        new_filename1 = base_ts_to_filename(base_ts + VIDEO_CLIP_EACH_SEC * 2)
        encode_process0 = subprocess.Popen([
            FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-y",
            "-hwaccel", "qsv", "-t", str(VIDEO_CLIP_EACH_SEC * 2 + (VIDEO_CLIP_OVERFLOW_SEC * 0.8)), "-i", orig_filename, "-vf",
            ",".join("subtitles=%s" % i for i in subtitles),
            "-c:a", "copy", "-c:v", "h264_qsv",
            "-f", "mp4", "-preset:v", "fast", "-profile:v", "high", "-level", "4.1",
            "-b:v", VIDEO_BITRATE, "-rc:v", "vbr", "-tune:v", "hq",
            "-qmin", "10", "-qmax", "32", "-crf", "16",
            # "-t", "10",
            new_filename0
        ])
        encode_process1 = subprocess.Popen([
            FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-y",
            "-ss", str(VIDEO_CLIP_EACH_SEC * 2), "-copyts", "-i", orig_filename, "-vf",
            ",".join("subtitles=%s" % i for i in subtitles),
            "-c:a", "copy", "-c:v", "h264_nvenc", "-ss", str(VIDEO_CLIP_EACH_SEC * 2),
            "-f", "mp4", "-preset:v", "fast", "-profile:v", "high", "-level", "4.1",
            "-b:v", VIDEO_BITRATE, "-rc:v", "vbr", "-tune:v", "hq",
            "-qmin", "10", "-qmax", "32", "-crf", "16",
            # "-t", "10",
            new_filename1
        ])
        encode_process1.wait()
        encode_process0.wait()
        return [new_filename0, new_filename1]
    elif duration > (VIDEO_CLIP_EACH_SEC * 1):
        # 至少也要能切2片，才用双GPU加速
        new_filename0 = base_ts_to_filename(base_ts, True)
        new_filename1 = base_ts_to_filename(base_ts + VIDEO_CLIP_EACH_SEC)
        encode_process0 = subprocess.Popen([
            FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-y",
            "-hwaccel", "qsv", "-t", str(VIDEO_CLIP_EACH_SEC + (VIDEO_CLIP_OVERFLOW_SEC * 0.8)), "-i", orig_filename, "-vf",
            ",".join("subtitles=%s" % i for i in subtitles),
            "-c:a", "copy", "-c:v", "h264_qsv",
            "-f", "mp4", "-preset:v", "fast", "-profile:v", "high", "-level", "4.1",
            "-b:v", VIDEO_BITRATE, "-rc:v", "vbr", "-tune:v", "hq",
            "-qmin", "10", "-qmax", "32", "-crf", "16",
            # "-t", "10",
            new_filename0
        ])
        encode_process1 = subprocess.Popen([
            FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-y",
            "-ss", str(VIDEO_CLIP_EACH_SEC), "-copyts", "-i", orig_filename, "-vf",
            ",".join("subtitles=%s" % i for i in subtitles),
            "-c:a", "copy", "-c:v", "h264_nvenc", "-ss", str(VIDEO_CLIP_EACH_SEC),
            "-f", "mp4", "-preset:v", "fast", "-profile:v", "high", "-level", "4.1",
            "-b:v", VIDEO_BITRATE, "-rc:v", "vbr", "-tune:v", "hq",
            "-qmin", "10", "-qmax", "32", "-crf", "16",
            # "-t", "10",
            new_filename1
        ])
        encode_process1.wait()
        encode_process0.wait()
        return [new_filename1]
    else:
        print("[-]VideoClip to short,", duration)
        print("[-]Fallback to normal")
        encode_video_with_subtitles(orig_filename, subtitles, new_filename)
        return [new_filename]


def encode_video_with_subtitles(orig_filename: str, subtitles: list[str], new_filename: str):
    if FFMPEG_USE_NVIDIA_GPU:
        print("[+]Use Nvidia NvEnc Acceleration")
        encode_process = subprocess.Popen([
            FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-y",
            "-i", orig_filename, "-vf",
            ",".join("subtitles=%s" % i for i in subtitles) + ",hwupload_cuda",
            "-c:a", "copy", "-c:v", "h264_nvenc",
            "-f", "mp4", "-preset:v", "fast", "-profile:v", "high", "-level", "4.1",
            "-b:v", VIDEO_BITRATE, "-rc:v", "vbr", "-tune:v", "hq",
            "-qmin", "10", "-qmax", "32", "-crf", "16",
            # "-t", "10",
            new_filename
        ], stdout=subprocess.PIPE)
    elif FFMPEG_USE_INTEL_GPU:
        if platform.system().lower() == "windows":
            print("[+]Use Intel QSV Acceleration")
            encode_process = subprocess.Popen([
                FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-y",
                "-hwaccel", "qsv", "-i", orig_filename, "-vf",
                ",".join("subtitles=%s" % i for i in subtitles),
                "-c:a", "copy", "-c:v", "h264_qsv",
                "-f", "mp4", "-preset:v", "fast", "-profile:v", "high", "-level", "4.1",
                "-b:v", VIDEO_BITRATE, "-rc:v", "vbr", "-tune:v", "hq",
                "-qmin", "10", "-qmax", "32", "-crf", "16",
                # "-t", "10",
                new_filename
            ], stdout=subprocess.PIPE)
        else:
            print("[+]Use Intel VAAPI Acceleration")
            encode_process = subprocess.Popen([
                FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-y",
                "-hwaccel", "vaapi", "-i", orig_filename, "-vf",
                ",".join("subtitles=%s" % i for i in subtitles) + ",hwupload",
                "-c:a", "copy", "-c:v", "h264_vaapi",
                "-f", "mp4", "-preset:v", "fast", "-profile:v", "high", "-level", "4.1",
                "-b:v", VIDEO_BITRATE, "-rc:v", "vbr", "-tune:v", "hq",
                "-qmin", "10", "-qmax", "32", "-crf", "16",
                # "-t", "10",
                new_filename
            ], stdout=subprocess.PIPE)
    else:
        print("[+]Use CPU Encode")
        encode_process = subprocess.Popen([
            FFMPEG_EXEC, "-hide_banner", "-progress", "-", "-loglevel", "error", "-y",
            "-i", orig_filename, "-vf",
            ",".join("subtitles=%s" % i for i in subtitles),
            "-c:a", "copy", "-c:v", "h264",
            "-f", "mp4", "-preset:v", "fast", "-profile:v", "high", "-level", "4.1",
            "-b:v", VIDEO_BITRATE, "-rc:v", "vbr",
            "-qmin", "10", "-qmax", "32", "-crf", "16",
            # "-t", "10",
            new_filename
        ], stdout=subprocess.PIPE)
    handle_ffmpeg_output(encode_process.stdout)
    return encode_process.wait()


def handle_ffmpeg_output(stderr: IO[bytes]) -> str:
    out_time = "0:0:0.0"
    while True:
        line = stderr.readline()
        if line == b"":
            break
        if line.strip() == b"progress=end":
            # 处理完毕
            break
        if line.startswith(b"out_time="):
            out_time = line.replace(b"out_time=", b"").decode().strip()
        if line.startswith(b"speed="):
            speed = line.replace(b"speed=", b"").decode().strip()
            print("Speed:", out_time, "@", speed)

    return out_time


def duration_str_to_float(duration_str) -> float:
    _duration = datetime.strptime(duration_str, "%H:%M:%S.%f") - datetime(1900, 1, 1)
    return _duration.total_seconds()


def quick_split_video(file):
    if not os.path.isfile(file):
        raise FileNotFoundError(file)
    file_name = os.path.split(file)[-1]
    _create_dt = os.path.splitext(file_name)[0]
    create_dt = datetime.strptime(_create_dt, "%Y%m%d_%H%M")
    _duration_str = get_video_real_duration(file)
    duration = duration_str_to_float(_duration_str)
    current_sec = 0
    while current_sec < duration:
        current_dt = (create_dt + timedelta(seconds=current_sec)).strftime("%Y%m%d_%H%M_")
        print("CUR_DT", current_dt)
        print("BIAS_T", current_sec)
        split_process = subprocess.Popen([
            FFMPEG_EXEC, "-y", "-hide_banner", "-progress", "-", "-loglevel", "error",
            "-ss", str(current_sec),
            "-i", file_name, "-c", "copy", "-f", "mp4",
            "-t", str(VIDEO_CLIP_EACH_SEC + VIDEO_CLIP_OVERFLOW_SEC),
            "{}.mp4".format(current_dt)
        ], stdout=subprocess.PIPE)
        handle_ffmpeg_output(split_process.stdout)
        current_sec += VIDEO_CLIP_EACH_SEC
